#include "shaunlib.h"
#include <fstream>

using namespace std;

int main(int argc, char* argv[])
{
    CRandom::srand();
    CArgs a(argc,argv);
    if (!a.find("--file")) error("Must specify --file my_file [--out my_output]...");
    string filename = a.value("--file");
    
    string outfilename = "downcode";
    if (a.find("--out")) outfilename=a.value("--out");
    
    CSample sample; 
    CDatfile dat;
    CMap locmap;

    dat.readDatfile(filename+".dat");
    locmap.readMapfile(filename+".map",dat);
    sample.readPedfile(filename+".ped",dat,locmap);

    
    vector< vector<string> > alleles;
    alleles.resize(sample.countLocus());

    for (int l=1; l<=sample.countLocus(); l++)
    {
	for (int f=1; f<=sample.countFamily(); f++)
	    for (int i=1; i<=sample.fam(f)->nInFamily(); i++)
	    {

		if (sample.fam(f)->ind(i)->loc(l).one != "0")
		{
		    
		    bool match = false;
		    
		    for (int a = 0 ; a < alleles[l-1].size(); a++)
			if (alleles[l-1][a] == sample.fam(f)->ind(i)->loc(l).one)
			    match = true;
		    
		    if (!match)
			alleles[l-1].push_back(sample.fam(f)->ind(i)->loc(l).one);
		    
		    
		    // second allele
		    match=false;
		    for (int a = 0 ; a < alleles[l-1].size(); a++)
			if (alleles[l-1][a] == sample.fam(f)->ind(i)->loc(l).two)
			    match = true;
		    
		    if (!match)
			alleles[l-1].push_back(sample.fam(f)->ind(i)->loc(l).two);
		}   
	    }
    }


    // output list ... 
    ofstream O;
    string f;
    
    // ALL file
    f = outfilename+".all";
    O.open(f.c_str());
    for (int l=1; l<=sample.countLocus(); l++)
    {
	if (alleles[l-1].size()<=2) O << "1\n";
	else O << alleles[l-1].size() << "\n";
    }
    O.close();

    // DAT file
    f = outfilename+".dat";
    O.open(f.c_str());
    // qt, bt, cov, then genotypes
    for (int l=1; l<=sample.countQTrait(); l++)
	O << "T\t" << sample.qt(l)->name << "\n";

    for (int l=1; l<=sample.countBTrait(); l++)
	O << "A\t" << sample.bt(l)->name << "\n";

    for (int l=1; l<=sample.countCovariate(); l++)
	O << "C\t" << sample.cov(l)->name << "\n";

    for (int l=1; l<=sample.countLocus(); l++)
    {
	if (alleles[l-1].size()<=2) O << "M\t"<<sample.loc(l)->name<<"\n";
	else for (int d=0; d<alleles[l-1].size(); d++)
	    O << "M\t" << sample.loc(l)->name << "_[" << alleles[l-1][d] << "]\n";

    }
    O.close();

    // MAP file
    f = outfilename+".map";
    O.open(f.c_str());
    for (int l=1; l<=sample.countLocus(); l++)
    {
	if (alleles[l-1].size()<=2) 
	    O << locmap.chr[sample.loc(l)->name] << "\t"
	      << sample.loc(l)->name             << "\t"
	      << locmap.cM[sample.loc(l)->name]  << "\t"
	      << locmap.bp[sample.loc(l)->name]  << "\n";
	else for (int d=0; d<alleles[l-1].size(); d++)
	    O << locmap.chr[sample.loc(l)->name] << "\t"
	      << sample.loc(l)->name << "_[" << alleles[l-1][d] << "]\t"
	      << locmap.cM[sample.loc(l)->name]  << "\t"
	      << locmap.bp[sample.loc(l)->name]  << "\n";	   
    }
    O.close();
    
    // PED file
    f = outfilename+".ped";
    O.open(f.c_str());
    
    for (int f=1; f<=sample.countFamily(); f++)
	for (int i=1; i<=sample.fam(f)->nInFamily(); i++)
	{
	    // Mandatory fields 
	    
	    O << sample.fam(f)->getID() << " "
	      << sample.fam(f)->ind(i)->getID() << " " ;

	    if (!sample.fam(f)->ind(i)->pat)
		O << "0 ";
	    else 
		O << sample.fam(f)->ind(i)->pat->getID() << " " ;
	    
	    if (!sample.fam(f)->ind(i)->mat)
		O << "0 ";
	    else 
		O << sample.fam(f)->ind(i)->mat->getID() << " " ;

	    O << sample.fam(f)->ind(i)->sex << " ";
	    
	    for (int t=1; t<=sample.countQTrait(); t++)
		O << sample.fam(f)->ind(i)->qt(t) << " ";
	    
	    for (int t=1; t<=sample.countBTrait(); t++)
		O << sample.fam(f)->ind(i)->bt(t)+1 << " ";
	    
	    for (int t=1; t<=sample.countCovariate(); t++)
		O << sample.fam(f)->ind(i)->cov(t) << " ";
	    
	    // write Genotyppes
	    for (int g=1; g<=sample.countLocus(); g++)
	    {
		if (alleles[g-1].size() <= 2) 
		{
		    // a normal SNP		    
    		    O << sample.fam(f)->ind(i)->loc(g).one << " "
			 << sample.fam(f)->ind(i)->loc(g).two << " ";
		}
		else
		{
		    // create a string of downcoded SNPs
		    for (int a=0;a<alleles[g-1].size(); a++)
		    {

			if (sample.fam(f)->ind(i)->loc(g).one == "0")
			    O << "0 ";
			else
			{
			    if (sample.fam(f)->ind(i)->loc(g).one == alleles[g-1][a])
				O << "2 ";
			    else 
				O << "1 ";
			}
			
			if (sample.fam(f)->ind(i)->loc(g).two == "0")
			    O << "0 ";
			else
			{
			    if (sample.fam(f)->ind(i)->loc(g).two == alleles[g-1][a])
				O << "2 ";
			    else 
				O << "1 ";
			}
			

		    }


		}
	    }
	    O << "\n";
	}
    O.close();
       
}
